/*
 * Unit tests for metafile functions
 *
 * Copyright (c) 2002 Dmitry Timoshkov
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>

#include "wine/test.h"
#include "winbase.h"
#include "wingdi.h"
#include "winerror.h"

#define STRUCTOFFSET(type,field) ((int)&((type *)0)->field)

static LOGFONTA orig_lf;
static BOOL emr_processed = FALSE;

static int CALLBACK emf_enum_proc(HDC hdc, HANDLETABLE *handle_table,
    const ENHMETARECORD *emr, int n_objs, LPARAM param)
{
    static int n_record;
    DWORD i;
    INT *dx, *orig_dx = (INT *)param;
    LOGFONTA device_lf;

    trace("hdc %p, emr->iType %ld, emr->nSize %ld, param %p\n",
           hdc, emr->iType, emr->nSize, (void *)param);

    PlayEnhMetaFileRecord(hdc, handle_table, emr, n_objs);

    switch (emr->iType)
    {
    case EMR_HEADER:
        n_record = 0;
        break;

    case EMR_EXTTEXTOUTA:
    {
        EMREXTTEXTOUTA *emr_ExtTextOutA = (EMREXTTEXTOUTA *)emr;
        dx = (INT *)((char *)emr + emr_ExtTextOutA->emrtext.offDx);

        ok(GetObjectA(GetCurrentObject(hdc, OBJ_FONT), sizeof(device_lf), &device_lf) == sizeof(device_lf),
           "GetObjectA error %ld", GetLastError());

        ok(!memcmp(&orig_lf, &device_lf, STRUCTOFFSET(LOGFONTA, lfFaceName)), "fonts don't match");
        ok(!lstrcmpA(orig_lf.lfFaceName, device_lf.lfFaceName), "font names don't match");

        for(i = 0; i < emr_ExtTextOutA->emrtext.nChars; i++)
        {
            ok(orig_dx[i] == dx[i], "pass %d: dx[%ld] (%d) didn't match %d",
                                     n_record, i, dx[i], orig_dx[i]);
        }
        n_record++;
        emr_processed = TRUE;
        break;
    }

    case EMR_EXTTEXTOUTW:
    {
        EMREXTTEXTOUTW *emr_ExtTextOutW = (EMREXTTEXTOUTW *)emr;
        dx = (INT *)((char *)emr + emr_ExtTextOutW->emrtext.offDx);

        ok(GetObjectA(GetCurrentObject(hdc, OBJ_FONT), sizeof(device_lf), &device_lf) == sizeof(device_lf),
           "GetObjectA error %ld", GetLastError());

        ok(!memcmp(&orig_lf, &device_lf, STRUCTOFFSET(LOGFONTA, lfFaceName)), "fonts don't match");
        ok(!lstrcmpA(orig_lf.lfFaceName, device_lf.lfFaceName), "font names don't match");

        for(i = 0; i < emr_ExtTextOutW->emrtext.nChars; i++)
        {
            ok(orig_dx[i] == dx[i], "pass %d: dx[%ld] (%d) didn't match %d",
                                     n_record, i, dx[i], orig_dx[i]);
        }
        n_record++;
        emr_processed = TRUE;
        break;
    }

    default:
        break;
    }

    return 1;
}

static void test_ExtTextOut(void)
{
    HDC hdcDisplay, hdcMetafile;
    HENHMETAFILE hMetafile;
    HFONT hFont;
    static const char text[] = "Simple text to test ExtTextOut on metafiles";
    INT i, dx[256];
    static const RECT rc = { 0, 0, 100, 100 };

    assert(sizeof(dx) >= lstrlenA(text));

    hdcDisplay = CreateDCA("display", NULL, NULL, NULL);
    ok(hdcDisplay != 0, "CreateDCA error %ld", GetLastError());

    trace("hdcDisplay %p\n", hdcDisplay);

    memset(&orig_lf, 0, sizeof(orig_lf));

    orig_lf.lfCharSet = ANSI_CHARSET;
    orig_lf.lfClipPrecision = CLIP_DEFAULT_PRECIS;
    orig_lf.lfWeight = FW_DONTCARE;
    orig_lf.lfHeight = 7;
    orig_lf.lfQuality = DEFAULT_QUALITY;
    lstrcpyA(orig_lf.lfFaceName, "Arial");
    hFont = CreateFontIndirectA(&orig_lf);
    ok(hFont != 0, "CreateFontIndirectA error %ld", GetLastError());

    hFont = SelectObject(hdcDisplay, hFont);
    for (i = 0; i < lstrlenA(text); i++)
    {
        ok(GetCharWidthA(hdcDisplay, text[i], text[i], &dx[i]),
           "GetCharWidthA error %ld", GetLastError());
    }
    hFont = SelectObject(hdcDisplay, hFont);

    hdcMetafile = CreateEnhMetaFileA(hdcDisplay, NULL, NULL, NULL);
    ok(hdcMetafile != 0, "CreateEnhMetaFileA error %ld", GetLastError());

    trace("hdcMetafile %p\n", hdcMetafile);

    ok(GetDeviceCaps(hdcMetafile, TECHNOLOGY) == DT_RASDISPLAY,
       "GetDeviceCaps(TECHNOLOGY) has to return DT_RASDISPLAY for a display based EMF");

    hFont = SelectObject(hdcMetafile, hFont);

    /* 1. pass NULL lpDx */
    ok(ExtTextOutA(hdcMetafile, 0, 0, 0, &rc, text, lstrlenA(text), NULL),
       "ExtTextOutA error %ld", GetLastError());

    /* 2. pass custom lpDx */
    ok(ExtTextOutA(hdcMetafile, 0, 20, 0, &rc, text, lstrlenA(text), dx),
       "ExtTextOutA error %ld", GetLastError());

    hFont = SelectObject(hdcMetafile, hFont);
    ok(DeleteObject(hFont), "DeleteObject error %ld", GetLastError());

    hMetafile = CloseEnhMetaFile(hdcMetafile);
    ok(hMetafile != 0, "CloseEnhMetaFile error %ld", GetLastError());

    ok(!DeleteDC(hdcMetafile) && GetLastError() == ERROR_INVALID_HANDLE,
       "CloseEnhMetaFile has to destroy metafile hdc");

    ok(PlayEnhMetaFile(hdcDisplay, hMetafile, &rc), "PlayEnhMetaFile error %ld", GetLastError());

    ok(EnumEnhMetaFile(hdcDisplay, hMetafile, emf_enum_proc, dx, &rc),
       "EnumEnhMetaFile error %ld", GetLastError());

    ok(emr_processed, "EnumEnhMetaFile couldn't find EMR_EXTTEXTOUTA or EMR_EXTTEXTOUTW record");

    ok(!EnumEnhMetaFile(hdcDisplay, hMetafile, emf_enum_proc, dx, NULL) && GetLastError() == ERROR_INVALID_PARAMETER,
       "A valid hdc has to require a valid rc");

    ok(DeleteEnhMetaFile(hMetafile), "DeleteEnhMetaFile error %ld", GetLastError());
    ok(DeleteDC(hdcDisplay), "DeleteDC error %ld", GetLastError());
}

START_TEST(metafile)
{
    /* make sure that user32 is really imported */
    LoadLibrary("user32");
    test_ExtTextOut();
}
